
"""
  Executes agglomerative hierarchical clustering on the taxpayers dataset [1]. scipy module for
  agglomerative hierarchical clustering is used. 

  Script executes agglomerative hierarchical clustering on data related to taxpayers that
  have features "Income","Spending","YearsWorking" and "NumChildren".
  All numerical features are used for clustering. The script uses the Ward method of calculating the 
  distance between clusters. The dendrogram is displayed. 
  
  The script also presents a trivial user interface that shows the clusters that occur when a specific
  dendrogram height is given. When specifying a height, the clusters, the cluster members along with the dendrogram
  cut by a line at the specified height is displayed.
  
  References:
     1) Taxpayers dataset, File taxpayers.csv.
     2) Clustering with Python — Hierarchical Clustering,
        https://anakin297.medium.com/clustering-with-python-hierarchical-clustering-a60688396945
 
  

  v0.8/mmt/Aug 2023

"""




# Import required libraries

import numpy as np

import pandas as pd
from pandas.api.types import is_numeric_dtype

from sklearn import preprocessing # for MinMax normalization

from scipy.cluster.hierarchy import linkage, cut_tree, dendrogram

from matplotlib import pyplot as plt




DATA_FILE = "taxpayers.csv"



# When tuning hyperparameter K, the maximum value
# of K to reach.
MAX_NUMBER_OF_CLUSTERS = 50

NUMBER_OF_CHARACTER_MARKS = 80



def DisplayDendrogram(lnkg, lbls, h=None):
    """Displays the dendrogram for a linkage matrix. May also display a cut point
       on the dendrogram. 

       Parameters:
          lnkg -- The linkage matrix representing the hierarchical clustering 
          lbls -- A range index object specifying the labels on the dendrogam
          h -- The floating point value representing the height of the dendrogram at
               which to draw a horizontal line indicating the clusters generated.
               A value of None will not draw any line.

          return -- float. The maximum hight at which clusters merge into a single cluster 
    """

    plt.figure(figsize=(10,6))
    denData = dendrogram(lnkg, leaf_rotation=90, leaf_font_size=10, labels=lbls)
    plt.title('Linkage Dendrogram')
    plt.xlabel('Data Points')
    plt.ylabel('Cluster Distance')

    # Plot a horizontal line indicating the clustering point
    if (h is not None):
        plt.axhline(y=h, color='r', linestyle='--')
        
    plt.show()

    # Return the maximum distance at which cluster join
    return(max(denData['dcoord'][-1]))






def printHeader(message = '', char='#', n=NUMBER_OF_CHARACTER_MARKS):
    """Displays a formatted header in the form of a box that separates
       strings shown during execution. Done to interpret/comprehend the
       displayed information better. 

       Parameters:
          message -- The message to display inside the header box
          char -- The character to use to create the box
          n -- The number of characters char to display for the box
               (modifies the box width only. Box height is fixed to 3 characters.)
       
    """
    
    if (message == ''):
        return

    print('')
    print('')
    print(NUMBER_OF_CHARACTER_MARKS*char)
    print(char, message)
    print(NUMBER_OF_CHARACTER_MARKS*char)
    print('')
    return





def main():

    # Read the data
    taxpayerData = pd.read_csv(DATA_FILE, sep=",", header=0, quotechar='"')
    #print(taxpayerData.dtypes)


    #################################################################################################
    # Preprocessing 
    #################################################################################################

    printHeader('Pre-processing')
    
    # Check for missing values. If there are missing values, remove them
    # NOTE1: taxpayerData.isna().sum() displays the number of missing values
    # for each column. The second .sum() adds the number of missing values in
    # each column producing a single number.
    # NOTE2: Could have called .dropna() directly. Doing it this way in case
    #        it is desired to see IF missing values exist.
    #
    if taxpayerData.isna().sum().sum() > 0 :
       taxpayerData =  taxpayerData.dropna()

    # Keep names; will be used as indexes
    indexes =  taxpayerData["Name"]


    # Normalize/scale data
    
    # Initialize scaler
    min_max_scaler = preprocessing.MinMaxScaler()
    taxpayerData = pd.DataFrame(min_max_scaler.fit_transform(taxpayerData.iloc[:, 1:5]), columns=taxpayerData.iloc[:, 1:5].columns)



    # Set as indexes the Names. Done so that names can be displayed on dendrogram's
    # x axis
    taxpayerData = taxpayerData.set_index(indexes)

    #################################################################################################
    # Hierarchical clustering (ward method) 
    #################################################################################################

    printHeader('Hierarchical clustering (ward method)')
    
    # Perform actual hierarchical clustering using the ward metric and Euclidean distance.
    # Other methods possible as well: single, complete etc
    hcLinkageMatrix = linkage(taxpayerData, method="ward", metric="euclidean")


    
    #################################################################################################
    # Dendrogram 
    #################################################################################################    

    printHeader('Displaying dendrogram')

    # Display hierarchical clustering as a dendrogram
    mxD = DisplayDendrogram(hcLinkageMatrix, taxpayerData.index)
    


    #################################################################################################
    # Cutting hierarchical tree at specific height 
    #################################################################################################

    printHeader('Cutting hierarchical tree (dendrogram). Showing interface')

    while (True):
       try: 
          h = float(input(f'Enter height <={"{:.3f}".format(mxD)} to cut tree and generate clusters (enter negative to quit):'))
          if h < 0:
             print('Terminating. ByeBye\n')
             break
            
       except Exception as e:
          print('Error. Wrong input.')
          continue
        
       # Cut dendrogram at given height h. This shows how easy it is - with hierarchical
       # clustering - to generate clusters without the need to re-execute the clustering
       # algorithm.
       # Here, for a given height, the clusters and their members will be displayed.
       # An alternative way is to specify the numbers of clusters to generate (not distance)
       # using the n_clusters instead of the height argument in cut_tree.
       #
       # This method of cutting the dendrogram can be used to generate the desired number of
       # clusters without the need to re-execute the clustering algorithm.
       # This way can also be used to find the optimal number of clusters to generate
       # when doing hierarchical clustering: see [2] for more details. 
       #
       # This can also be used as a preprocessing step prior to executing K-means to
       # get the optimal number of K to cluster the data. This is an alternative
       # to the Elbow and Silhouette methods.
       clusterLabels = cut_tree(hcLinkageMatrix, height=h).flatten()
       
       print('Total of ', len(set(clusterLabels)), 'clusters')
       print('Displaying clusters and their contents:')
       
       for i in range(max(clusterLabels)+1):
           # See to which cluster each taxpayer was assigned to.
           # Use cluster labels to match data and the cluster they belong to.
           nms = taxpayerData.iloc[list(np.where(clusterLabels==i)[0]), :]
           print(f'\tCluster {i}: ', end='')
           print(', '.join(list(nms.index.astype(str))) )
       
       # Display dendrogram with cut point at height h
       DisplayDendrogram(hcLinkageMatrix, taxpayerData.index, h)
        
    return(0)






# Main guard
# See: https://stackoverflow.com/questions/19578308/what-is-the-benefit-of-using-main-method-in-python
#
# Script starts execution from here.
if __name__ == "__main__":
    main()

